# BASIC 2 line-numbering too
# (c) 2022 Tom Wilson wilsontp@gmail.com
# License: cc-by 
# free to use and remix, attribution with my name and email address is required for any derivative work
#
# This program adds line numbers to non-numbered BASIC programs
# GOTO and GOSUB use labels. A label starts with @ and uses letters, numbers, or _
# Label must be the first symbol on a line. Text after label is treated as a comment
#   @test infinite loop
#   GOTO @test
# becomes
#   100 REM INFINITE LOOP
#   110 GOTO 100
#
# @@ will substitue the current line number
#   GET A$:IF A$="" GOTO @@
#
# labels are not case sensitive and will be converted to upper-case internally

bas_lines = {}
bas_labels = {}
bas_errors = []

# add_numbers 
# in_text: list of str, un-numbered BASIC program
# creates bas_lines {} with {line_number : Program Text}
# creates bas_labels {} with {label_name : label_line}
def add_numbers(in_text):
    global bas_lines;  bas_lines = {0:"REM CONVERTED BY BASIC2"} # Line Number,Line Text 
    global bas_labels; bas_labels = {}

    ln = 10
    for line in in_text:
        l=line.strip()
        if len(l) == 0:
            continue

        # save label address
        if l[0] == '@':
            bas_labels[l.upper()] = ln 
            bas_lines[ln] = "REM " + l.upper()
        # output line 
        else:
            bas_lines[ln] = l
        ln += 10

# returns True if c is a word character
# a word character is a letter, number, @, or _ symbol
# returns false for any other character
def is_word(c):
    if c == None:
        return False
    if c >= 'A' and c <= 'Z':
        return True
    if c >= 'a' and c <= 'z':
        return True 
    if c >= '0' and c <= '9':
        return True 
    if c == '@':
        return True 
    if c == '_':
        return True 
    return False 

# gets a word in text, starting at pos
# a word is any series of letters, nubmers, and _ followed by a non-word character or end of string.
def get_word(text, pos):
    i = pos
    wl = 0
    while i < len(text) and is_word(text[i]):
        i += 1        
    return text[pos:i]    

# converts all instances of @label in a program
# to the equivalent line number 
def convert_labels():
    global bas_lines, bas_labels, bas_errors

    for k,v in bas_lines.items():
        if v[0:3].upper() == "REM":
            continue
        
        # start by looking for a label and fall though if none exists
        x=v.find("@")
        new_text = v
        while x>=0:
            label = get_word(new_text,x)
            if label == "@@":
                new_text = new_text.replace(label, str(k))
                bas_lines[k] = new_text                 
            elif label in bas_labels:
                line_num = bas_labels[label]
                new_text = new_text.replace(label, str(line_num))
                bas_lines[k] = new_text                 
            else:
                bas_errors.append(" Label not found on " + str(k) + " : " + label)
                break 
            x = new_text.find("@")
            

# prints the BASIC program
def print_program():
    global bas_lines, bas_labels, bas_errors
    print("BASIC LISTING")
    for k,v in bas_lines.items():
        print(k,v)
    
    print()
    print("LABELS")
    for k,v in bas_labels.items():
        print(k,v)

def print_errors():
    global bas_errors
    if len(bas_errors) > 0:
        print()
        print("ERRORS")
        for err in bas_errors:
            print(err)
    else:
        print("No errors")

# prints the BASIC program
def save_bas(filename):
    outf=open(filename + ".bas",'w')

    global bas_lines, bas_labels, bas_errors
    if len(bas_errors) > 0:
        outf.write("Errors reported. See end of file\n")
    
    for k,v in bas_lines.items():
        outf.write(str(k))
        outf.write(" ")
        outf.write(v)
        outf.write("\n")

    if len(bas_errors) > 0:
        outf.write("\nERRORS\n")
        for err in bas_errors:
            outf.write(err)
            outf.write("\n")
    outf.close()

# File to read
filename = "Morse Code"
fi=open(filename + ".ba2",'r')
b2_text = fi.readlines() 
fi.close()

bas_errors = []
add_numbers(b2_text)
convert_labels()
#print_program()
print_errors()

# File to write 
save_bas(filename)
# save_sym(filename)
# save_lst(filename)
